"""
Implementation of dslibrary that makes REST calls.
"""
import json
import re
import io
import time

try:
    import requests
except ImportError:
    requests = None

from ..exceptions import DSLibraryException, DSLibraryCommunicationError
from ..metadata import Metadata
from ..front import DSLibrary
from ..utils import filechunker
from ..utils import dbconn


PTN_MODE = re.compile(r'^[rwa]b?$')


class DSLibraryViaREST(DSLibrary):
    """
    A model can send all its dslibrary requests over REST and have them handled elsewhere.
    """
    def __init__(self, url: str, token=None, spec: dict=None, auth_header_name: str="X-DSLibrary-Token"):
        super(DSLibraryViaREST, self).__init__(spec=spec)
        if not url.endswith("/"):
            url += "/"
        self._url = url
        self._headers = {auth_header_name: token} if token else {}
        self._session = requests.session() if requests else None
        self._meta = None
        self._context = None
        self._timeouts = (90, 10)

    def _do_comm(self, method: str, path: str, params: dict=None, data=None, as_json: bool=True):
        """
        All HTTP communication goes through here.
        """
        if not self._session:
            # TODO let's use urllib instead of requests when requests is not available
            raise DSLibraryException("The 'requests' package is required to use dslibrary in this mode.")
        try:
            resp = self._session.request(
                method, self._url + path, params=params, data=data, headers=self._headers, timeout=self._timeouts
            )
            resp.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise DSLibraryCommunicationError(*err.args)
        if as_json:
            try:
                return resp.json()
            except ValueError as err:
                raise DSLibraryCommunicationError(f"JSON encoding error: {err}")
        return resp

    def _fetch_context(self):
        if self._context is None:
            self._context = self._do_comm("get", "context") or {}

    def get_metadata(self) -> Metadata:
        """
        Metadata can be supplied by the remote endpoint.  We could check for local metadata here, but it gives the
        remote controller more flexibility if we let it be in charge of this.
        """
        if not self._meta:
            self._fetch_context()
            self._meta = Metadata(**self._context.get("metadata") or {})
        return self._meta

    def get_parameters(self):
        """
        All the processing of parameters is delegated to the remote.
        """
        self._fetch_context()
        return self._context.get("parameters") or {}

    def _open_stream(self, category: str, url: str, mode: str, filename: str=None, params: dict=None, chunk_size=5000000):
        """
        General purpose read/write file streams over HTTP.
        """
        if not PTN_MODE.match(mode):
            raise ValueError("invalid mode: %s" % mode)
        if "r" in mode:
            def read_chunk(start, end):
                resp = self._do_comm("get", category, params={"resource_name": url, "byte_range": json.dumps([start, end]), **(params or {})}, as_json=False)
                return resp.content if "b" in mode else resp.text
            return filechunker.ChunkedFileReader(filename, reader=read_chunk, mode=mode, chunk_size=chunk_size)
        else:
            def write_chunk(content, append, hint=None):
                if ("b" in mode) == isinstance(content, str):
                    raise TypeError("incorrect data type for mode=%s" % mode)
                resp = self._do_comm("put", category, params={"resource_name": url, "append": append, "hint": hint}, data=content)
                return resp.get("upload_id")
            return filechunker.ChunkedFileWriter(filename, writer=write_chunk, mode=mode)

    def _opener(self, path: str, mode: str, **kwargs) -> io.IOBase:
        return self._open_stream("resources", path, mode, filename=path)

    def open_run_data(self, filename: str, mode: str='rb') -> io.RawIOBase:
        """
        Shared pipeline data has a different endpoint.
        """
        return self._open_stream("run_data", filename, mode, filename=filename)

    def next_scoring_request(self, timeout: float=None) -> (dict, None):
        """
        Scoring requests come from a remote endpoint.
        """
        timeout = timeout or 60
        t_timeout = time.time() + timeout
        interval = min(10, timeout/3)
        while True:
            resp = self._do_comm("get", "scoring-requests", params={"timeout": interval}, as_json=True)
            if resp.get("shutdown"):
                raise StopIteration
            if "request" in resp:
                return resp["request"] or {}
            if time.time() > t_timeout:
                return

    def send_score(self, score):
        """
        Scoring responses go to a remote endpoint.
        """
        self._do_comm("post", "score", params={"value": score})

    def get_sql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
        """
        SQL conversation is 'chunked across REST', so to speak.
        """
        if not isinstance(resource_name, str):
            raise ValueError(f"get_db_connection(): expected str for 'connection_name', got {type(resource_name)}")
        params = dict(resource_name=resource_name)

        def read(operation, parameters):
            # FIXME re-establish this upstream
            '''
            # pandas makes this specific query to verify a table does not already exist (pandas.DataFrame.to_sql())
            if "select name from sqlite_master" in operation.lower():
                # convert to standard SQL
                operation = "select distinct table_name as name from INFORMATION_SCHEMA.TABLES where table_name=?"
                # try to qualify based on database
                if database:
                    operation += f" and table_schema='{database}'"
            '''
            resp = self._do_comm(
                "get", "db", params=dict(**params, sql=operation, parameters=json.dumps(parameters))
            )
            cols = resp[0]
            rows = resp[1:]
            more = None
            if rows and isinstance(rows[-1], str):
                more = rows.pop(-1)
            descr = [[col, None, None, None, None, None] for col in cols]
            return descr, rows, more

        def read_more(chunk):
            resp = None
            for _ in range(15):
                resp = self._do_comm("get", "db", params=dict(**params, more=chunk, max_wait=60))
                if resp:
                    break
            if not resp:
                raise DSLibraryCommunicationError("timeout waiting for results")
            rows = resp[1:]
            more = None
            if rows and isinstance(rows[-1], str):
                more = rows.pop(-1)
            return rows, more

        def write(operation, parameters):
            self._do_comm("post", "db", data=dict(**params, sql=operation, parameters=parameters))

        return dbconn.Connection(read, write if for_write else None, read_more=read_more)



import unittest
import pandas

from dslibrary.utils.connect import _process_params, dataframe_to_sql


class TestConnect(unittest.TestCase):
    def test__process_params(self):
        r = _process_params("mysql://u:p@host:1234/database", extra=1)
        assert r == {
            'host': 'host',
            'port': 1234,
            'database': 'database',
            'username': 'u',
            'password': 'p',
            'extra': 1
        }, r
        r = _process_params("xyz://u@host")
        assert r == {
            'host': 'host',
            'port': None,
            'database': '',
            'username': 'u',
            'password': ''
        }, r
        r = _process_params("sqlite:path/to/file")
        assert r == {
            'host': '',
            'port': None,
            'database': 'path/to/file',
            'username': '',
            'password': ''
        }, r
        r = _process_params("x://host/db", database="db2")
        assert r == {
            'host': 'host',
            'port': None,
            'database': 'db2',
            'username': '',
            'password': ''
        }, r

    def test_dataframe_to_sql(self):
        df = pandas.DataFrame({"x": [1, 2, 3], "y": [1.0, 2.0, 4.0], "z": ["one", "two", "three"]})
        sqls = list(dataframe_to_sql(df, "t1"))
        self.assertEqual(sqls, [
            ('DROP TABLE IF EXISTS t1', []),
            ('CREATE TABLE IF NOT EXISTS t1 (x INTEGER, y DOUBLE, z VARCHAR(8))', []),
            ('INSERT INTO t1 (x, y, z) VALUES (%s, %s, %s), (%s, %s, %s), (%s, %s, %s)', [1, 1.0, 'one', 2, 2.0, 'two', 3, 4.0, 'three'])
        ])



"""
Custom protocol handlers.
"""
import os

from .utils.instantiation import instantiate_class_and_args


class ProtocolHandler(object):
    """
    Base class for protocol-specific handlers.
    """
    def open_resource(self, url: str, mode: str, **kwargs):
        """
        Open a r/w stream to a specific file-like object.
        :param url:         Which file.
        :param mode:        r, rb, w, wb, a, ab
        :param kwargs:      Custom arguments.
        :return:            A file-like object.
        """

    def get_system_connection(self, system_type, uri: str, for_write: bool=False, **kwargs):
        """
        Create a connection to external systems of all types.
        :param system_type:     filesystem, sql, nosql, timeseries, ...
        :param uri:             Base URL for the remote system.
        :param for_write:       Requests write access.
        :param kwargs:          Arguments to help open the connection.
        :return:                See classes in 'engine_intf'.
        """


def load_protocol_handler(url_or_protocol: str, env: dict=None) -> (ProtocolHandler, None):
    """
    Load a protocol handler, given a protocol, or a URL containing a protocol.

    :param url_or_protocol:     Protocol, or a URL containing a protocol, i.e. "myprotocol://hostname/etc."
    :param env:                 Alternate environement.  Uses system environment if not specified.
    """
    if env is None:
        env = os.environ
    # extract protocol
    if "://" in url_or_protocol:
        protocol = url_or_protocol.split("://")[0]
    elif "/" not in url_or_protocol and ":" not in url_or_protocol and " " not in url_or_protocol:
        protocol = url_or_protocol
    else:
        return
    # look up environment variable to see if the given protocol has a special handler
    env_name = f"DSLIBRARY_PROTO_{protocol.upper()}"
    env_val = env.get(env_name)
    if not env_val:
        return
    # look up the referenced handler class and instantiate it
    return instantiate_class_and_args(env_val, ProtocolHandler)



import unittest
import json

from dslibrary.metadata import Metadata


class TestMetadata(unittest.TestCase):
    def test_json_round_trip(self):
        m = {
            "uri": "u",
            "entry_points": {
                "e1": {
                    "parameters": {"x": {"type": "float", "default": 2}},
                    "command": "python abc.py",
                    "inputs": {"i1": {"columns": [{"name": "a", "type": "integer"}]}},
                    "outputs": {"o1": {"columns": [{"name": "b", "type": "string"}]}}
                }
            }
        }
        j1 = Metadata.from_json(m).to_json()
        j2 = Metadata.from_json(j1).to_json()
        assert j1 == j2



"""
Exception classes.
"""


class DSLibraryException(Exception):
    """
    General errors.
    """


class DSLibraryDataFormatException(DSLibraryException):
    """
    Problems with data
    """
    def __init__(self, message: str, offset: int=None):
        super(DSLibraryDataFormatException, self).__init__(message, offset)
        self.message = message
        self.offset = offset


class DSLibraryCommunicationError(DSLibraryException):
    """
    Communication errors.
    """





import unittest
import mock
import pandas
import tempfile

from dslibrary.run_models import ModelRunner


class TestRunToMLFlow(unittest.TestCase):
    """
    MLFlow can be made the target of some or all output operations.
    """
    def test_log_text(self):
        """
        The compatibility functions like log_text() just call the equivalent mlflow method.
        """
        runner = ModelRunner(uri="uri", entry_point="main", mlflow={"all": True})
        def my_model(dsl):
            log = []
            with mock.patch("mlflow.log_text", lambda txt, fn: log.append((txt, fn))):
                dsl.log_text("hello", "text1")
            assert log == [('hello', 'text1')]
        runner.run_method(my_model)

    def test_write_df_to_log_artifact(self):
        """
        Writing of data frames uses filename appropriate format and sends via log_artifact().
        """
        runner = ModelRunner(uri="uri", entry_point="main", mlflow={"all": True})
        def my_model(dsl):
            df = pandas.DataFrame({"x": [1, 2], "y": [3, 4]})
            log = []
            def log_artifact(fn, rmt):
                with open(fn) as f_r:
                    data = f_r.read()
                log.append((data, rmt))
            with mock.patch("mlflow.log_artifact", log_artifact):
                dsl.write_resource("data.csv", df)
            assert log == [('x,y\n1,3\n2,4\n', 'data.csv')]
        runner.run_method(my_model)

    def test_write_df_with_format_options(self):
        """
        Writing of data frames uses filename appropriate format and sends via log_artifact().
        """
        runner = ModelRunner(uri="uri", entry_point="main", mlflow={"all": True})
        def my_model(dsl):
            df = pandas.DataFrame({"x": [1, 2], "y": [3, 4]})
            log = []
            def log_artifact(fn, rmt):
                with open(fn) as f_r:
                    data = f_r.read()
                log.append((data, rmt))
            with mock.patch("mlflow.log_artifact", log_artifact):
                dsl.write_resource("data.csv", df)
            self.assertEqual(log, [('x\ty\n1\t3\n2\t4\n', 'data.csv')])
        runner.set_output("data.csv", "", format_options={"sep": "\t"})
        runner.run_method(my_model)

    def test_mapped_input(self):
        """
        Input mapping is unchanged by the mlflow settings.
        """
        with tempfile.NamedTemporaryFile(suffix=".csv") as f_tmp:
            runner = ModelRunner(uri="uri", entry_point="main", mlflow=True)
            def my_model(dsl):
                df = dsl.load_dataframe("data.csv")
                assert list(df.x) == [1, 2, 3]
            f_tmp.write(b"x\n1\n2\n3\n")
            f_tmp.flush()
            runner.set_input("data.csv", f_tmp.name)
            runner.run_method(my_model)



"""
Run a model.
"""
import json
import inspect
import os
import typing
import subprocess
import sys
import tempfile
import pandas

from pexpect import split_command_line

import dslibrary
from dslibrary import ENV_DSLIBRARY_TOKEN
from dslibrary.front import DSLibraryException, ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, METRICS_ALIAS
from dslibrary.metadata import Metadata


class ModelRunner(object):
    """
    Each instance sets up for one run of one model.
    """
    def __init__(self, uri: str=None, user: str=None, run_id: str=None, project_root: str=None, entry_point: str = "main", mlflow: (bool, dict)=None, parameters: dict=None):
        """
        The constructor provides some context which is optional when directly invoking local code, and required
        when invoking code remotely.

        :param uri:             Unique identifier for the model.
        :param user:            Which user is running the model.
        :param run_id:          Run ID needs to be specified when running in MLFlow.
        :param entry_point:     Which entry point is being called.
        :param mlflow:          MLFlow-related settings.  True = send all to mlflow, {} = send some calls to mlflow.
        """
        self.uri = uri or ""
        self.user = user or ""
        self.run_id = run_id or ""
        self.project_root = project_root
        self.entry_point = entry_point
        self.mlflow = mlflow if isinstance(mlflow, dict) else {"all": True} if mlflow is True else {}
        self.inputs = {}
        self.outputs = {}
        self.data = {}
        self.parameters = parameters or {}
        self.code_paths = []

    def _to_absolute(self, uri: str):
        """
        Convert local paths to absolute paths except when they match a certain pattern.  "~" signifies the
        model's root path and can be used to access built-in model data.
        """
        # TODO consider self.project_root
        # TODO test this
        # blank > means no change to supplied name
        if not uri:
            return ""
        # no changes to URIs
        if ":" in uri:
            return uri
        if uri.startswith("~/"):
            # specifically point to files within the sandbox
            return "./" + uri[1:]
        if uri.startswith("/"):
            # already absolute
            return uri
        # convert to absolute path
        return os.path.abspath(uri)

    def set_input(self, name: str, uri: str="", data=None, **kwargs):
        """
        Specify where data will come from for a particular named input.
        :param name:    Name of input.
        :param uri:     A path to a local file, or the URI of remote data.  Or, a URI specifying a sql or nosql data
                        source.
        :param data:    Specific in-memory data can be supplied when running locally.
        :param kwargs:  Additional parameters to support the various data sources, following fsspec for file-like
                        sources.
        """
        self.inputs[name] = {"uri": self._to_absolute(uri), **kwargs}
        if data is not None:
            self.data[name] = data
        return self

    def set_output(self, name: str, uri: str="", capture: bool=False, **kwargs):
        """
        Specify where data should go for a particular named output.

        A format can be chosen by specifying "format".
          csv or tab -- remaining arguments are sent to pandas.to_csv()
          json - remaining arguments are sent to pandas.to_json()
          etc.

        :param name:    Name of output.
        :param uri:     See set_input()
        :param capture: Output content from write_resource() can be captured in memory.
        :param kwargs:  See set_input()
        """
        spec = {"uri": self._to_absolute(uri), **kwargs}
        self.outputs[name] = spec
        if capture:
            spec["capture"] = True
        return self

    def set_parameter(self, name: str, value):
        """
        Specify a value for one of the parameters.
        """
        self.parameters[name] = value
        return self

    def set_parameters(self, params: dict):
        """
        Specify values for multiple parameters.
        """
        self.parameters.update(params)
        return self

    def add_code_path(self, path: str):
        """
        Point to a location where import modules might exist.
        :param path:   Path to a folder containing python code to import from your model.
        """
        if path not in self.code_paths:
            self.code_paths.append(path)
        return self

    def send_metrics_to(self, uri: str=None, mlflow: bool=None, capture: bool=False, **kwargs):
        """
        Determine where metrics will be stored (for read and write).  Parameters are equivalent to set_output(), in
        that metrics can be sent to a particular location.

        :param uri:         A URI or filename.
        :param mlflow:      Specify True to have the model log its metrics through MLFlow.
        :param kwargs:      Additional arguments.
        """
        if mlflow is not None:
            self.mlflow["metrics"] = bool(mlflow)
        self.inputs[METRICS_ALIAS] = {**kwargs}
        self.outputs[METRICS_ALIAS] = {**kwargs}
        if uri:
            self.inputs[METRICS_ALIAS]["uri"] = uri
            self.outputs[METRICS_ALIAS]["uri"] = uri
        if capture:
            self.outputs[METRICS_ALIAS]["capture"] = True
        return self

    def _generate_spec(self, local: bool=False):
        spec = {
            "uri": self.uri,
            "user": self.user,
            "run_id": self.run_id,
            "entry_point": self.entry_point,
            "inputs": self.inputs,
            "outputs": self.outputs,
            "parameters": self.parameters,
            "code_paths": self.code_paths
        }
        if self.data and local:
            spec["data"] = self.data
        if self.mlflow:
            spec["mlflow"] = self.mlflow
        return spec

    def run_method(self, method: typing.Callable, path: str=None):
        """
        As long as 'method()' uses dslibrary.instance(), and all calls through this method are single threaded, you can
        use this method, and it is very efficient.
        """
        # map parameters to method signature
        call_kwargs = {}
        sig = inspect.signature(method)
        sig_kwargs = None
        remaining_params = dict(self.parameters)
        for p_name, param in sig.parameters.items():
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                sig_kwargs = True
                continue
            if p_name in remaining_params:
                value = remaining_params.pop(p_name)
                if param.annotation and param.annotation is not inspect.Parameter.empty:
                    value = param.annotation(value)
                call_kwargs[p_name] = value
        if sig_kwargs:
            call_kwargs.update(remaining_params)
        # as long as the method lets us send a custom dslibrary instance we can do that
        if "dsl" in sig.parameters or "dslibrary" in sig.parameters:
            run_env = self.prepare_run_env(local_path=path, local=True)
            dsl = dslibrary.instance(run_env)
            call_kwargs["dsl" if "dsl" in sig.parameters else "dslibrary"] = dsl
            retval = method(**call_kwargs)
        else:
            # otherwise we assume method() calls dslibrary.instance()
            # - this approach is not thread safe
            # TODO is this approach even worth supporting???
            run_env = self.prepare_run_env(local_path=path, local=False)
            dsl = dslibrary
            orig_env = {ENV_DSLIBRARY_TARGET: os.environ.get(ENV_DSLIBRARY_TARGET), ENV_DSLIBRARY_SPEC: os.environ.get(ENV_DSLIBRARY_SPEC)}
            try:
                # store specification
                for k, v in run_env.items():
                    os.environ[k] = v or ""
                # the method makes calls to dslibrary, following rules set above
                retval = method(**call_kwargs)
            finally:
                # restore environment
                for k, v in orig_env.items():
                    os.environ[k] = v or ""
        # we return captured data, in case there is any, and it replaces the method return value
        if hasattr(dsl, "_spec"):
            captured = dsl._spec.get("capture")
            if captured:
                retval = captured
        return retval

    def run_local(self, path: str, entry_point: str=None, extra_env: dict=None):
        """
        Execute a python, R or notebook based model in a subprocess.
        """
        # if project_root is specified, 'path' is relative to it, by default
        path = path or ""
        if self.project_root and not path.startswith("/"):
            path = os.path.join(self.project_root, path)
        # verify target exists, split apart path
        if not os.path.exists(path):
            raise DSLibraryException(f"Path not found: {path}")
        # if a folder has been specified we have to look up 'entry_point' and run its official command
        cmd = None
        if not path or os.path.isdir(path):
            folder = self.project_root or path or "."
            meta = Metadata.from_folder(folder)
            entry_point = entry_point or self.entry_point or "main"
            if entry_point not in meta.entry_points:
                raise DSLibraryException(f"Entry point {entry_point} not found in model at {path}")
            cmd_str = meta.entry_points[entry_point].command
            if not cmd_str:
                raise DSLibraryException(f"Entry point {entry_point} for {path} does not define a command")
            cmd = split_command_line(cmd_str)
            # insert venv here as well (if 'python' is the command used)
            if cmd and cmd[0] == "python":
                venv_py = self.find_venv_python(folder)
                if venv_py:
                    cmd[0] = venv_py
        else:
            folder = self.project_root or os.path.split(path)[0] or "."
        # if we're not using metadata we have to infer how to execute the code
        if not cmd:
            cmd = self.infer_command_from_path(path)
        # convert in-memory data to files
        temp_files = []
        for resource_name, data in self.data.items():
            with tempfile.NamedTemporaryFile(delete=False) as f_w:
                if hasattr(data, "to_csv"):
                    data.to_csv(f_w, sep="\t", index=False)
                # TODO don't clobber self.inputs, this should be strictly temporary
                self.inputs[resource_name]["uri"] = f_w.name
                self.inputs[resource_name]["format"] = "csv"
                self.inputs[resource_name]["format_options"] = {"delimiter": "\t"}
                temp_files.append(f_w.name)
        # when capturing outputs, redirect to temporary file
        output_capture_files = {}
        for resource_name, spec in self.outputs.items():
            if spec.get("capture"):
                tmp_fn = tempfile.NamedTemporaryFile(delete=True).name
                temp_files.append(tmp_fn)
                # TODO don't clobber this either
                spec["capture"] = False
                spec["uri"] = output_capture_files[resource_name] = tmp_fn
                spec["format"] = "csv"
                spec["format_options"] = {"sep": "\t"}
        # work out environment variables
        env = {
            **self.prepare_run_env(local_path=folder),
            **(extra_env or {})
        }
        # call the model
        # TODO report failures in a more elegant way
        subprocess.run(cmd, cwd=folder, env=env, check=True)
        # capture outputs
        outputs = None
        for resource_name, filename in output_capture_files.items():
            if not os.path.exists(filename):
                continue
            if not outputs:
                outputs = {}
            # TODO output might not be a dataframe
            outputs[resource_name] = pandas.read_csv(filename, sep="\t")
        # convert captured metrics from dataframe to dict
        if outputs and METRICS_ALIAS in outputs:
            metrics = {}
            for _, row in outputs[METRICS_ALIAS].iterrows():
                metrics[row["name"]] = row.value
            outputs[METRICS_ALIAS] = metrics
        # delete temporary files for self.data and captured outputs
        for temp_file in temp_files:
            if os.path.exists(temp_file):
                os.remove(temp_file)
        # return captured output
        return outputs

    @staticmethod
    def find_venv_python(folder: str):
        venv_py = os.path.join(folder, "venv", "bin", "python")
        if not os.path.exists(venv_py):
            return
        return venv_py

    @staticmethod
    def infer_command_from_path(path: str):
        """
        When we only know the name of a source file we have to work out how to execute it.

        Used by run_local().
        """
        folder, fn = os.path.split(path)
        _, f_ext = os.path.splitext(fn)
        f_ext = f_ext.lower()
        # detect virtual environment
        # TODO 'path' is not the best place to get 'folder'
        venv_py = ModelRunner.find_venv_python(folder)
        # if we're not using metadata we have to infer how to execute the code
        if f_ext == ".py":
            cmd = [venv_py or sys.executable, path]
        elif f_ext == ".r":
            # TODO detect venv for R
            cmd = ["RScript", path]
        elif f_ext == ".ipynb":
            # FIXME apply venv_py here!
            cmd = ["nbconvert", "--to", "notebook", "--execute", "--inplace", path]
        else:
            raise DSLibraryException(f"unrecognized executable extension: {f_ext} in {fn}")
        return cmd

    def prepare_run_env(self, rest_url: str=None, shared_volume: str=None, local_path: str=None, rest_access_token: str=None, local: bool=False) -> dict:
        """
        Prepare environment variables for a run which encapsulate all of the settings generated by calls to this
        class.  Data science code running anywhere, with these environment variables set, can use dslibrary to read
        and write all of its data without having to know any details about data location, credentials or format.

        Three approaches are provided:
          * rest_url and rest_access_token - fill this in to have the target code use a REST service for data access.
          * shared_volume - fill this in to use a shared volume (i.e. a sidecar) for data access.
          * local_path - this option causes all inputs and outputs to default to files in the indicated folder, and
              there is no delegation of data access, it is all performed by the process running the target code.

        In a high security scenario which isolates the data science code from the credentials and other data source
        details:
          * Send the ENV_DSLIBRARY_SPEC value to a secure service (a REST service indicated by rest_uri, or a sidecar
            sharing the volume 'shared_volume' with the target code).
          * Send the other environment variables to the target code.
          * The target code calls dslibrary methods, which communicate with the data service, and the data service
            performs all reads and writes based on the environment data it has been sent.

        The Kubernetes sidecar approach involves running a container alongside your target code's container, and having
        them both mount the same shared, ephemeral volume.  The appropriate environment variables are set for each, and
        voila, the target code is able to perform all its data access without having exposed any sensitive information.

        :param rest_url:            URL of REST service.
        :param shared_volume:       Path to shared volume.
        :param local_path:          Local path.
        :param rest_access_token:   Access token to secure communications.
        :return:  Environment variables to set, as a {}.
        """
        env = {}
        if shared_volume:
            env[ENV_DSLIBRARY_TARGET] = f"volume:{shared_volume}"
        elif rest_url:
            env[ENV_DSLIBRARY_TARGET] = rest_url
            if rest_access_token:
                env[ENV_DSLIBRARY_TOKEN] = rest_access_token
        elif local_path:
            env[ENV_DSLIBRARY_TARGET] = f"local:{local_path}"
        spec = self._generate_spec(local)
        env[ENV_DSLIBRARY_SPEC] = spec if local else json.dumps(spec)
        return env

    def run_mlflow(self, model_uri: str):
        """
        Use mlflow.run() to execute an MLFlow model.

        The trick will be how to set the environment variables in the remote model such that the interception takes
        place and sends data back to us.  Could initially only be supported when running locally with an environment
        variable.
        """
        # TODO code me
        # TODO I don't see a way to have MLFlow deliver extra environment variables!



import unittest
import mock

from dslibrary.transport.to_rest import DSLibraryViaREST


class TestToRest(unittest.TestCase):

    def test_metadata_and_parameters(self):
        log = []
        class MyDSL(DSLibraryViaREST):
            def _do_comm(self, method: str, path: str, params: dict=None, data=None, as_json: bool=True):
                log.append((method, path, params))
                return {"metadata": {}, "parameters": {"x": 1}}
        dsl = MyDSL("URL")
        assert dsl.get_parameter("x") == 1

    # TODO resources
    # TODO sql



"""
Communication via shared volume.

In order to truly 'lock down' user-supplied code, and elimate the security risks associated with giving it access
to the network, or in providing it credentials for communicating with external REST services, all of mmlibrary's
communications can be sent to a folder shared with a sidecar (or another mounted volume).  The sidecar then transfers
the requests and responses to a REST API.
"""
import base64
import json
import os
import threading
import time

from ..exceptions import DSLibraryCommunicationError
from .to_rest import DSLibraryViaREST


class DSLibraryViaVolume(DSLibraryViaREST):
    def __init__(self, volume: str, spec: dict=None, _env: dict=None):
        super(DSLibraryViaVolume, self).__init__(url="", spec=spec)
        self._volume = volume
        self._request_no = 0
        self._lock = threading.Lock()

    class Response(object):
        def __init__(self, content: bytes):
            self.content = content

        @property
        def text(self):
            return self.content.decode("utf-8")

    def _do_comm(self, method: str, path: str, params: dict=None, data=None, as_json: bool=True):
        """
        Creation of a file in the folder is equivalent to making a REST call.  The response is returned in a file with
        a related filename.
        """
        t_timeout = time.time() + self._timeouts[1]
        with self._lock:
            self._request_no += 1
            fn_rqst = os.path.join(self._volume, f"rqst_{self._request_no}.json")
            fn_resp = fn_rqst + ".response"
            rqst = {
                "method": method, "path": path, "params": params, "data": data, "timeout": t_timeout, "as_json": as_json
            }
            if isinstance(data, (bytes, bytearray)):
                rqst["data"] = base64.b64encode(data).decode("utf-8")
                rqst["data_format"] = "base64"
            rq_content = json.dumps(rqst)
            with open(fn_rqst, 'w') as f_w:
                f_w.write(rq_content)
            # now we wait for a response...
            try:
                while True:
                    time.sleep(0.01)
                    if time.time() > t_timeout:
                        raise DSLibraryCommunicationError(f"timeout waiting for {path} request")
                    if os.path.exists(fn_resp):
                        try:
                            with open(fn_resp, 'rb') as f_r:
                                if as_json:
                                    resp = json.load(f_r)
                                else:
                                    # NOTE: the burden is on the other side to deliver the entire response 'atomically'
                                    resp = DSLibraryViaVolume.Response(f_r.read())
                        except ValueError:
                            continue
                        return resp
            finally:
                os.remove(fn_rqst)
                if os.path.exists(fn_resp):
                    os.remove(fn_resp)


